import datetime
import difflib
import os
from inspect import isabstract
from typing import Any, List, Mapping, Optional, Set, Tuple, Type

import ee  # type: ignore
from requests.structures import CaseInsensitiveDict

config_path = os.path.expanduser("~/.config/taskee.ini")


def initialize_earthengine() -> None:
    """Initialize the Earth Engine API."""
    try:
        ee.Initialize()
    except ee.EEException:
        ee.Authenticate()
        ee.Initialize()


def _get_case_insensitive_close_matches(
    word: str, possibilities: List[str], n: int = 3, cutoff: float = 0.6
) -> List[str]:
    """A case-insensitive wrapper around difflib.get_close_matches.
    Parameters
    ----------
    word : str
        A string for which close matches are desired.
    possibilites : List[str]
        A list of strings against which to match word.
    n : int, default 3
        The maximum number of close matches to return. n must be > 0.
    cutoff : float, default 0.6
        Possibilities that don't score at least that similar to word are ignored.

    Returns
    -------
    List[str] : The best (no more than n) matches among the possibilities are returned in
        a list, sorted by similarity score, most similar first.
    """
    lower_matches = difflib.get_close_matches(
        word.lower(), [p.lower() for p in possibilities], n, cutoff
    )
    return [p for p in possibilities if p.lower() in lower_matches]


def _all_subclasses(cls: Type[Any]) -> Set[Type[Any]]:
    """Recursively find all subclasses of a given class."""
    return set(cls.__subclasses__()).union(
        [s for c in cls.__subclasses__() for s in _all_subclasses(c)]
    )


def _list_subclasses(superclass: Type[Any]) -> Mapping[str, Type[Any]]:
    """List all non-abstract subclasses of a given superclass. Return as a dictionary mapping the
    subclass name to the class. This is recursive, so sub-subclasses will also be returned.

    Parameters
    ----------
    superclass : Type[Any]
        The superclass to list subclasses of.

    Returns
    -------
    Dict[str, Type[Any]]
        A dictionary mapping the subclass name to the class.
    """
    return CaseInsensitiveDict(
        {
            cls.__name__: cls
            for cls in _all_subclasses(superclass)
            if not isabstract(cls)
        }
    )


def _get_subclasses(names: Tuple[str, ...], superclass: Type[Any]) -> Set[Type[Any]]:
    """Retrieve a set of subclasses of a given superclass.

    Parameters
    ----------
    names : Tuple[str, ...]
        A tuple of subclass names to retrieve from the superclass.
    superclass : Type[Any]
        The superclass to retrieve subclasses of.

    Returns
    -------
    Set[Type[Any]]
        A set of subclasses of the superclass.
    """
    options = _list_subclasses(superclass)
    keys = list(options.keys())

    if "all" in [name.lower() for name in names if isinstance(name, str)]:
        return set(options.values())

    selected = []

    for name in names:
        try:
            selected.append(options[name])
        except KeyError:
            close_matches = _get_case_insensitive_close_matches(name, keys, n=3)
            hint = " Close matches: {}.".format(close_matches) if close_matches else ""

            raise AttributeError(
                f'"{name}" is not a supported {superclass.__name__} type. Choose from {keys}.{hint}'
            )

    return set(selected)


def _millis_to_datetime(
    millis: str, tz: Optional[datetime.timezone] = None
) -> datetime.datetime:
    """Convert a timestamp in milliseconds (e.g. from Earth Engine) to a datetime object."""
    return datetime.datetime.fromtimestamp(int(millis) / 1000.0, tz=tz)


def _datetime_to_millis(dt: datetime.datetime) -> int:
    """Convert a datetime to a timestamp in milliseconds"""
    return int(dt.timestamp() * 1000)



CANCELLED = "CANCELLED"
CANCEL_REQUESTED = "CANCEL_REQUESTED"
RUNNING = "RUNNING"
COMPLETED = "COMPLETED"
FAILED = "FAILED"
READY = "READY"

FINISHED = [COMPLETED, FAILED, CANCELLED]
ACTIVE = [READY, RUNNING]
UNSUCCESSFUL = [FAILED, CANCELLED]
ALL = [CANCELLED, CANCEL_REQUESTED, RUNNING, COMPLETED, FAILED, READY]



import datetime
import logging
import time
from typing import Tuple

from rich.logging import RichHandler
from rich.status import Status

from taskee import events
from taskee.cli.styles import get_style
from taskee.taskee import Taskee

logging.basicConfig(
    format="%(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
    handlers=[RichHandler(show_level=True, show_path=False, markup=True)],
)

logger = logging.getLogger("taskee")


def start(
    t: Taskee,
    watch_for: Tuple[str, ...] = ("error", "completed", "failed"),
    interval_minutes: float = 5.0,
) -> None:
    """Run an indefinite logger. This handles scheduling of Earth Engine updates and
    logs events as they occur.
    """
    logger.setLevel("INFO")

    last_checked = time.time()
    interval_seconds = interval_minutes * 60.0

    watch_events = events.get_events(watch_for)

    while True:
        elapsed = time.time() - last_checked

        if elapsed > interval_seconds:
            with Status(f"[yellow]Updating tasks...", spinner="bouncingBar"):
                new_events = t._update(watch_events)
                last_checked = time.time()

            if len(new_events) > 1:
                new_events = tuple(sorted(new_events, key=lambda event: event.time))

            for event in new_events:
                muted_style = "[dim]" if event.__class__ not in watch_events else ""
                style = get_style(event.__class__)
                logger.info(
                    f"[{style.color}]{style.emoji} {event.__class__.__name__}[/]:"
                    f" {muted_style}{event.message}"
                )

        else:
            delta = datetime.timedelta(seconds=interval_seconds - elapsed)
            next_update = datetime.datetime.now() + delta

            with Status(
                f"[yellow]Next update at {next_update:%H:%M:%S}...",
                spinner="bouncingBar",
            ):
                time.sleep(delta.total_seconds())



from abc import ABC, abstractmethod
from typing import Mapping, Set, Tuple, Type

from taskee.utils import _get_subclasses, _list_subclasses


class Notifier(ABC):
    @abstractmethod
    def send(self, title: str, message: str) -> None:
        pass


def list_notifiers() -> Mapping[str, Type[Notifier]]:
    """List all Notifier subclasses. Return as a dictionary mapping the subclass name to the
    class.

    Returns
    -------
    Dict[str, Type[Notifier]]
        A dictionary mapping the Notifier name to the class.
    """
    return _list_subclasses(Notifier)


def get_notifiers(names: Tuple[str, ...]) -> Set[Type[Notifier]]:
    """Retrieve a set of subclasses of Notifier.

    Parameters
    ----------
    names : Tuple[str, ...]
        A list of names of Notifiers to retrieve.

    Returns
    -------
    Set[Type[Notifier]]
        A set of Notifiers.
    """
    return _get_subclasses(names, Notifier)



from typing import Set, Tuple, Type

import ee  # type: ignore

from taskee import events, utils
from taskee.dispatcher import Dispatcher
from taskee.tasks import TaskManager


class Taskee:
    """The Taskee object is the primary interface to the taskee library. It connects the task manager, which
    tracks tasks and events, to the dispatcher, which distributes notifications. The Taskee object is the
    only direct point of contact with Earth Engine."""

    def __init__(self, notifiers: Tuple[str, ...] = ("native",)):

        """
        Parameters
        ----------
        notifiers : Tuple[str, ...]
            Notifier names for handling notifications.
        """
        utils.initialize_earthengine()
        self.manager = TaskManager(ee.data.getTaskList())
        self.dispatcher = Dispatcher(notifiers)

    def _update(self, watch_for: Set[Type[events.Event]]) -> Tuple[events.Event, ...]:
        """Update tasks and return any events that occured. Dispatch notifications for events of interest."""
        self.manager.update(ee.data.getTaskList())

        new_events = self.manager.events

        for event in new_events:
            if isinstance(event, tuple(watch_for)):
                self.dispatcher.notify(event.title, event.message)

        return new_events



import configparser
from typing import TYPE_CHECKING, Union

from requests.exceptions import ConnectionError
from rich.prompt import Prompt  # type: ignore

from taskee.notifiers.notifier import Notifier
from taskee.utils import config_path

if TYPE_CHECKING:
    import pushbullet  # type: ignore


class Pushbullet(Notifier):
    def __init__(self) -> None:
        self.pb = initialize_pushbullet()

    def send(self, title: str, message: str) -> None:
        push = self.pb.push_note(title, message)


def initialize_pushbullet() -> "pushbullet.Pushbullet":
    """Initialize the Pushbullet API and return a Pushbullet object."""
    try:
        import pushbullet
    except ImportError:
        raise ImportError(
            "The `pushbullet` package must be installed to use the Pushbullet notifier."
            " Run `pip install pushbullet.py` to install."
        )

    api_key = _get_stored_pushbullet_key(config_path)

    store_key = False
    pb: Union[str, None] = None
    while pb is None:
        try:
            pb = pushbullet.Pushbullet(api_key)
        except pushbullet.errors.InvalidKeyError:
            api_key = _request_pushbullet_key()
            store_key = True
        except ConnectionError:
            raise ConnectionError(
                "Failed to connect to Pushbullet! Please make sure you are connected to"
                " internet and try again."
            )

    if store_key:
        _store_pushbullet_key(api_key, config_path)

    return pb


def _get_stored_pushbullet_key(path: str) -> str:
    """Get the stored Pushbullet API key from a config file, if it exists. If the file,
    section, or key don't exist, return None.

    Parameters
    ----------
    path : str
        The path to the config file.
    """
    config = configparser.ConfigParser()
    config.read(path)

    try:
        return config["Pushbullet"]["api_key"]
    except KeyError:
        return ""


def _request_pushbullet_key() -> str:
    """Request a Pushbullet API key from the user."""
    apikey = Prompt.ask("Enter your [yellow bold]Pushbullet[/] API key to continue")
    return apikey


def _store_pushbullet_key(key: str, path: str) -> None:
    """Store the Pushbullet API key in the config file. If the config file does not exist,
    it will be created.

    Parameters
    ----------
    key : str
        The Pushbullet API key to store.
    path : str
        The path to the config file.
    """
    config = configparser.ConfigParser()
    config.read(path)

    try:
        config["Pushbullet"]["api_key"] = key
    # If Pushbullet section doesn't exist, create it
    except KeyError:
        config["Pushbullet"] = {"api_key": key}

    with open(path, "w") as dst:
        config.write(dst)



from typing import Tuple

from taskee.cli.commands import log


def test(notifiers: Tuple[str, ...]) -> None:
    from taskee.dispatcher import Dispatcher

    log.logger.setLevel("INFO")
    dispatcher = Dispatcher(notifiers)

    dispatcher.notify(
        title="Notification Test",
        message="If you receive this notification, taskee is working!",
    )
    log.logger.info("Notification sent!")



import datetime
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Mapping, Set, Tuple, Type

import humanize  # type: ignore

from taskee import utils

if TYPE_CHECKING:
    from taskee.tasks import Task


class Event(ABC):
    title = "Generic Event"

    def __init__(self) -> None:
        self.time = datetime.datetime.now()

    @property
    @abstractmethod
    def message(self) -> str:
        return ""


class TaskEvent(Event, ABC):
    """A Task Event is a type of event originating from an Earth Engine task."""

    def __init__(self, task: "Task"):
        self.task = task
        super().__init__()


class Error(Event):
    """An Error event occurs when taskee crashes."""

    title = "Oops!"
    message = "Something went wrong and taskee needs to be restarted."


class Failed(TaskEvent):
    """A Failed event occurs when a task fails to complete."""

    title = "Task Failed"

    @property
    def message(self) -> str:
        error = self.task.error_message
        elapsed = humanize.naturaldelta(self.task.time_elapsed)
        return f"Task '{self.task.description}' failed with error '{error}' after {elapsed}."


class Completed(TaskEvent):
    """A Completed event occurs when a task completes successfully."""

    title = "Task Completed"

    @property
    def message(self) -> str:
        elapsed = humanize.naturaldelta(self.task.time_elapsed)
        return f"Task '{self.task.description}' completed successfully! It ran for {elapsed}."


class Created(TaskEvent):
    """A Created event occurs when a new task is created."""

    title = "Task Created"

    @property
    def message(self) -> str:
        return f"Task '{self.task.description}' was created."


class Attempted(TaskEvent):
    """An Attempted event occurs when an attempt fails and a new attempt beings."""

    title = "Attempt Failed"

    @property
    def message(self) -> str:
        n = self.task._status["attempt"]
        return f"Task '{self.task.description}' attempt {n - 1} failed."


class Cancelled(TaskEvent):
    """A Cancelled event occurs when a task is cancelled by the user."""

    title = "Task Cancelled"

    @property
    def message(self) -> str:
        return f"Task '{self.task.description}' was cancelled."


class Started(TaskEvent):
    """A Started event occurs when a ready task begins running."""

    title = "Task Started"

    @property
    def message(self) -> str:
        return f"Task '{self.task.description}' has started processing."


def list_events() -> Mapping[str, Type["Event"]]:
    """List all Event subclasses. Return as a dictionary mapping the subclass name to the
    class.

    Returns
    -------
    Dict[str, Type[Event]]
        A dictionary mapping the Event name to the class.
    """
    return utils._list_subclasses(Event)


def get_events(names: Tuple[str, ...]) -> Set[Type[Event]]:
    """Retrieve a set of subclasses of Event.

    Parameters
    ----------
    names : Tuple[str, ...]
        A list of names of Events to retrieve.

    Returns
    -------
    Set[Type[Any]]
        A set of Events.
    """
    return utils._get_subclasses(names, Event)



import datetime
from typing import Tuple

import ee  # type: ignore
import humanize  # type: ignore
import rich
from rich import box
from rich.status import Status
from rich.table import Table

from taskee import states
from taskee.cli.styles import get_style
from taskee.tasks import Task
from taskee.utils import initialize_earthengine


def tasks() -> None:
    initialize_earthengine()

    with Status("Retrieving tasks from Earth Engine...", spinner="bouncingBar"):    
        tasks = [Task(task) for task in ee.data.getTaskList()]

        rich.print(create_task_table(tuple(tasks)))


def create_task_table(tasks: Tuple[Task, ...]) -> Table:
    """Create a table of tasks."""
    t = Table(
        title="[bold bright_green]Tasks",
        box=box.SIMPLE_HEAD,
        header_style="bright_green",
        expand=True,
    )

    t.add_column("State", justify="right")
    t.add_column("Description", justify="left")
    t.add_column("Created", justify="right")
    t.add_column("Elapsed", justify="right")

    for task in tasks:
        state_style = get_style(task.state)
        dim_style = "[dim]" if task.state not in states.ACTIVE else ""
        time_created = humanize.naturaltime(task.time_created)
        time_elapsed = humanize.naturaldelta(task.time_elapsed)

        t.add_row(
            f"[{state_style.color}]{task.state}[/] {state_style.emoji}",
            f"{dim_style}{task.description}",
            f"{dim_style}{time_created}",
            f"{dim_style}{time_elapsed}",
        )

    return t



import os

import notifypy  # type: ignore

from taskee.notifiers.notifier import Notifier


class Native(Notifier):
    def send(self, title: str, message: str) -> None:
        notification = notifypy.Notify()
        notification.application_name = "taskee"
        notification.icon = os.path.join(
            os.path.dirname(__file__), "icons", "taskee.png"
        )
        notification.title = title
        notification.message = message
        notification.send()



from typing import TYPE_CHECKING, Any, Mapping, Type, Union

from taskee import events, states

COLOR_SUCCESS = "green"
COLOR_ERROR = "red"
COLOR_INFO = "cyan"
COLOR_WARNING = "yellow"

if TYPE_CHECKING:
    from taskee.events import Event


class Style:
    def __init__(self, color: str, emoji: str = ""):
        self.color = color
        self.emoji = emoji


styles: Mapping[Any, Style] = {
    events.Error: Style(color=COLOR_ERROR, emoji=":exclamation:"),
    events.Failed: Style(color=COLOR_ERROR, emoji=":fire:"),
    events.Completed: Style(color=COLOR_SUCCESS, emoji=":evergreen_tree:"),
    events.Created: Style(color=COLOR_INFO, emoji=":seedling:"),
    events.Started: Style(color=COLOR_INFO, emoji=":herb:"),
    events.Attempted: Style(color=COLOR_WARNING, emoji=":fallen_leaf:"),
    events.Cancelled: Style(color=COLOR_ERROR, emoji=":axe:"),
    states.CANCEL_REQUESTED: Style(color=COLOR_ERROR, emoji=":triangular_flag:"),
    states.CANCELLED: Style(color=COLOR_ERROR, emoji=":prohibited:"),
    states.COMPLETED: Style(color=COLOR_SUCCESS, emoji=":white_heavy_check_mark:"),
    states.FAILED: Style(color=COLOR_ERROR, emoji=":x:"),
    states.READY: Style(color=COLOR_INFO, emoji=":hourglass:"),
    states.RUNNING: Style(color=COLOR_INFO, emoji=":wrench:"),
}


def get_style(obj: Union[str, Type["Event"]]) -> Style:
    """Retrieve the Style for a given object or class, such as an Event class or a state string."""
    try:
        return styles[obj]
    except KeyError:
        raise KeyError(f"'{obj}' does not have a registered style.")



from typing import Tuple

import click  # type: ignore

from taskee import events
from taskee.cli.commands import dashboard, log, tasks, test
from taskee.notifiers import notifier
from taskee.taskee import Taskee

version = "0.0.3"
modes = {"log": log.start, "dashboard": dashboard.start}


@click.group()
@click.version_option(version, prog_name="taskee")
def taskee() -> None:
    """Monitor Earth Engine tasks and send notifications when they change states.

    \b
    Examples
        $ taskee test
        $ taskee tasks
        $ taskee start log
        $ taskee start dashboard failed completed -n pushbullet -i 0.5
    """
    return


@taskee.command(name="start")
@click.argument("mode", nargs=1, type=click.Choice(choices=modes.keys()))
@click.argument(
    "watch_for",
    nargs=-1,
    type=click.Choice(
        choices=list(events.list_events().keys()) + ["all"], case_sensitive=False
    ),
)
@click.option(
    "notifiers",
    "-n",
    "--notifier",
    default=("native",),
    multiple=True,
    type=click.Choice(
        list(notifier.list_notifiers().keys()) + ["all"], case_sensitive=False
    ),
    help="One or more notifiers to run.",
)
@click.option(
    "interval_mins",
    "-i",
    "--interval_mins",
    default=5.0,
    help="Minutes between queries to Earth Engine for task updates.",
)
def start_command(
    mode: str,
    watch_for: Tuple[str, ...],
    notifiers: Tuple[str, ...],
    interval_mins: float,
) -> None:
    """Start running the notification system. Select a mode (default native)
    and one or more event types to watch for (or all).

    \b
    Examples
        $ taskee start dashboard failed completed -n pushbullet -i 5
        $ taskee start log all
    """
    if len(watch_for) == 0:
        watch_for = ("completed", "failed", "error")

    mode_func = modes[mode]
    t = Taskee(notifiers=notifiers)

    try:
        mode_func(t, watch_for=watch_for, interval_minutes=interval_mins)
    except Exception as e:
        if "error" in [event.lower() for event in watch_for]:
            event = events.Error()
            t.dispatcher.notify(event.title, event.message)
        raise e
    except KeyboardInterrupt:
        return


@taskee.command(name="tasks")
def tasks_command() -> None:
    """Display a table of current Earth Engine tasks."""
    tasks.tasks()


@taskee.command(name="test", short_help="Send test notifications.")
@click.option(
    "notifiers",
    "-n",
    "--notifier",
    default=("native",),
    multiple=True,
    type=click.Choice(
        list(notifier.list_notifiers().keys()) + ["all"], case_sensitive=False
    ),
    help="One or more notifiers to test.",
)
def test_command(notifiers: Tuple[str, ...]) -> None:
    """
    Send test notifications to selected notifiers (default native).

    \b
    Examples
        $ taskee test -n all
    """
    test.test(notifiers)


if __name__ == "__main__":
    taskee()



from typing import Tuple

from taskee.notifiers.notifier import get_notifiers


class Dispatcher:
    def __init__(self, notifiers: Tuple[str, ...]):
        selected_notifiers = get_notifiers(notifiers)
        self.notifiers = [notifier() for notifier in selected_notifiers]

    def notify(self, title: str, message: str) -> None:
        """Send notifications to all registered notifiers."""
        for notifier in self.notifiers:
            notifier.send(title, message)



from setuptools import find_packages, setup  # type: ignore

version = "0.0.3"

requirements = ["earthengine-api", "rich", "humanize", "notify-py", "requests", "click"]
test_requirements = ["pytest"]
dev_requirements = [
    "pre-commit",
    "black",
    "isort",
    "bumpversion",
    "twine",
    "mypy",
    "types-requests",
] + test_requirements
extras_require = {"dev": dev_requirements, "test": test_requirements}

with open("README.md", encoding="utf8") as readme_file:
    readme = readme_file.read()

setup(
    name="taskee",
    version=version,
    description="Notifications for Earth Engine tasks.",
    long_description=readme + "\n\n",
    keywords="earth-engine, notifications, tasks, cli, command-line",
    author="Aaron Zuspan",
    author_email="aa.zuspan@gmail.com",
    url="https://github.com/aazuspan/taskee",
    license_files=("LICENSE",),
    license="GPLv3+",
    install_requires=requirements,
    tests_require=test_requirements,
    extras_require=extras_require,
    python_requires=">=3.7",
    test_suite="tests",
    packages=find_packages(),
    long_description_content_type="text/markdown",
    include_package_data=True,
    entry_points="""
        [console_scripts]
        taskee=taskee.cli.cli:taskee
    """,
)



import datetime
from typing import Dict, List, Tuple, Union

from taskee import events, states
from taskee.utils import _millis_to_datetime


class Task:
    """Wrapper class around a persistent Earth Engine task."""

    def __init__(self, obj: Dict):
        self._status = obj
        self.id = obj["id"]
        self.description = obj["description"]
        self.time_created = _millis_to_datetime(obj["creation_timestamp_ms"])
        self.event: Union[events.Event, None] = events.Created(self)

    def __str__(self) -> str:
        return f"<{self.id}>: {self.description} [{self.state}]"

    @property
    def error_message(self) -> str:
        if self.state == states.FAILED:
            return self._status["error_message"]
        return ""

    @property
    def state(self) -> str:
        """Get the last state of the task."""
        return self._status["state"]

    @property
    def time_updated(self) -> datetime.datetime:
        return _millis_to_datetime(self._status["update_timestamp_ms"])

    @property
    def time_elapsed(self) -> datetime.timedelta:
        """Return the time elapsed between the task creation and the last update."""
        return self.time_updated - self.time_created

    def update(self, new_status: Dict) -> None:
        """Update the status of the task and record any changed attributes as events."""
        self.event = self._parse_event(new_status)
        self._status = new_status

    def _parse_event(self, new_status: Dict) -> Union[None, events.Event]:
        """Take the updated status dictionary and identify if an Event occured."""
        event: Union[None, events.Event] = None

        old_state = self._status["state"]
        new_state = new_status["state"]

        try:
            old_attempt, new_attempt = self._status["attempt"], new_status["attempt"]
        # During some states, tasks will not have an attempt value
        except KeyError:
            old_attempt, new_attempt = 0, 0

        if old_state != new_state:
            if new_state == states.RUNNING:
                event = events.Started(self)
            elif new_state == states.CANCELLED:
                event = events.Cancelled(self)
            elif new_state == states.FAILED:
                event = events.Failed(self)
            elif new_state == states.COMPLETED:
                event = events.Completed(self)

        if old_attempt != new_attempt:
            event = events.Attempted(self)

        return event


class TaskManager:
    """Manager class for handling all Earth Engine tasks."""

    def __init__(self, tasks: List[Dict]):
        self._tasks: Dict = {}
        self.update(tasks)

        # The initial set of tasks should not register Created events.
        # There's a better way to handle this, but for now I'm just manually suppressing those events.
        for task in self._tasks.values():
            task.event = None

    @property
    def events(self) -> Tuple[events.Event, ...]:
        """Retrieve active events from all tasks, sorted by time."""
        task_events = [task.event for task in self.tasks if task.event is not None]
        if len(task_events) > 1:
            task_events = sorted(task_events, key=lambda event: event.time)
        return tuple(task_events)

    @property
    def tasks(self) -> Tuple[Task, ...]:
        return tuple(self._tasks.values())

    def update(self, task_list: List[Dict]) -> None:
        """Update all tasks. Existing tasks will be updated and new tasks will be added to the manager."""
        current_tasks = self._tasks.keys()

        for task in task_list:
            task_id = task["id"]

            # Store new tasks
            if task_id not in current_tasks:
                new_task = Task(task)
                self._tasks[task_id] = new_task

            # Update existing tasks
            else:
                self._tasks[task_id].update(task)

        self._sort_tasks()

    def _sort_tasks(self) -> None:
        """Sort the tasks by placing active tasks first and then sorting by their creation date."""
        self._tasks = {
            k: v
            for k, v in sorted(
                self._tasks.items(),
                key=lambda x: [x[1].state in states.ACTIVE, x[1].time_created],
                reverse=True,
            )
        }



import time
from collections import deque
from typing import TYPE_CHECKING, Deque, Optional, Tuple

import humanize  # type: ignore
from rich import box
from rich.layout import Layout
from rich.live import Live
from rich.panel import Panel
from rich.progress_bar import ProgressBar
from rich.table import Table
from rich.text import Text

from taskee import events, states
from taskee.cli.styles import get_style
from taskee.taskee import Taskee

REFRESH_SECONDS = 1
MAX_ROWS = 20
TABLE_HEADER_HEIGHT = 6
BOX_STYLE = box.SIMPLE_HEAD

if TYPE_CHECKING:
    from taskee.events import Event


class Dashboard:
    def __init__(
        self,
        t: Taskee,
        watch_for: Tuple[str, ...] = ("error", "completed", "failed"),
        interval_minutes: float = 5.0,
    ):
        self.event_log: Deque["Event"] = deque(maxlen=MAX_ROWS)
        self.last_checked = time.time()

        self.t = t
        self.watch_events = events.get_events(watch_for)
        self.interval_seconds = interval_minutes * 60.0

        self.layout = self.create_layout()
        self.window = Panel(
            self.layout,
            title="[bold white]taskee",
            border_style="bright_black",
            height=36,
        )

        # Initialize the dashboard before we start Live so we don't render a preview layout
        self.update_display()

    @property
    def elapsed(self) -> float:
        """Return the time elapsed since the last event update"""
        return time.time() - self.last_checked

    @property
    def time_remaining(self) -> float:
        """Return the time remaining until the next event update"""
        return self.interval_seconds - self.elapsed

    def run(self) -> None:
        """Run the dashboard indefinitely."""
        with Live(self.window):
            while True:
                if self.elapsed > self.interval_seconds:
                    self.update_events()

                self.update_display()
                time.sleep(REFRESH_SECONDS)

    def update_events(self) -> None:
        """Update Earth Engine tasks and store new events."""
        self.t._update(self.watch_events)

        new_events = self.t.manager.events
        for event in new_events:
            self.event_log.appendleft(event)

        self.last_checked = time.time()

    def update_display(self) -> None:
        """Update the dasboard display."""
        self.layout["header"].update(self.create_header())
        self.layout["progress"].update(self.create_progress())

        task_table, event_table = self.create_tables()
        self.layout["tasks"].update(task_table)
        self.layout["events"].update(event_table)
        self.layout["tasks"].size = task_table.row_count + TABLE_HEADER_HEIGHT
        self.layout["events"].size = event_table.row_count + TABLE_HEADER_HEIGHT

    def create_layout(self) -> Layout:
        """Create the dashboard Layout"""
        layout = Layout()

        layout.split(
            Layout(name="header", size=1),
            Layout(name="progress", size=2),
            Layout(name="events", minimum_size=4),
            Layout(name="tasks", minimum_size=4),
        )

        layout["progress"].update(self.create_progress())
        layout["header"].update(self.create_header())

        return layout

    def create_header(self) -> Table:
        """Create the dashboard Header showing the update time and controls."""
        grid = Table.grid(expand=True)
        grid.add_column(justify="left")
        grid.add_column(justify="right")
        grid.add_row(
            "[italic]Next update in"
            f" {humanize.naturaldelta(self.time_remaining)}...[/]",
            Text("Press CTRL + C to exit...", style="dim"),
        )

        return grid

    def create_progress(self) -> ProgressBar:
        return ProgressBar(
            total=self.interval_seconds,
            completed=self.interval_seconds - self.time_remaining,
            complete_style="bright_yellow",
        )

    def create_tables(self) -> Tuple[Table, Table]:
        n_tasks = MAX_ROWS - min(max(len(self.event_log), 1), MAX_ROWS // 2)
        n_events = MAX_ROWS - n_tasks

        task_table = self.create_task_table(n_tasks)
        event_table = self.create_event_table(n_events)

        return task_table, event_table

    def create_task_table(self, max_tasks: Optional[int] = None) -> Table:
        """Create a table of tasks."""
        t = Table(
            title="[bold bright_green]Tasks",
            box=BOX_STYLE,
            header_style="bright_green",
            expand=True,
        )

        t.add_column("State", justify="right")
        t.add_column("Description", justify="left")
        t.add_column("Created", justify="right")
        t.add_column("Elapsed", justify="right")

        for task in self.t.manager.tasks[:max_tasks]:
            state_style = get_style(task.state)
            dim_style = "[dim]" if task.state not in states.ACTIVE else ""
            time_created = humanize.naturaltime(task.time_created)
            time_elapsed = humanize.naturaldelta(task.time_elapsed)

            t.add_row(
                f"[{state_style.color}]{task.state}[/] {state_style.emoji}",
                f"{dim_style}{task.description}",
                f"{dim_style}{time_created}",
                f"{dim_style}{time_elapsed}",
            )

        return t

    def create_event_table(self, max_events: Optional[int] = None) -> Table:
        """Create a table of events."""
        t = Table(
            title="[bold bright_blue]Events",
            box=BOX_STYLE,
            header_style="bright_blue",
            expand=True,
        )

        t.add_column("Event", justify="right")
        t.add_column("Message", justify="left")
        t.add_column("Time", justify="right")

        for event in tuple(self.event_log)[:max_events]:
            event_time = humanize.naturaltime(event.time)
            event_style = get_style(event.__class__)
            event_name = event.__class__.__name__
            muted_style = "[dim]" if event.__class__ not in self.watch_events else ""
            t.add_row(
                f"[{event_style.color}]{event_name}[/] {event_style.emoji}",
                f"{muted_style}{event.message}",
                event_time,
            )
        if len(self.event_log) == 0:
            t.add_row("", Text("No events yet..."), "", style="dim italic bright_black")

        return t


def start(
    t: Taskee,
    watch_for: Tuple[str, ...] = ("error", "completed", "failed"),
    interval_minutes: float = 5.0,
) -> None:
    """Run an indefinite dashboard. This handles scheduling of Earth Engine updates and
    runs a live-updating dashboard of tasks and events as they occur.
    """
    dashboard = Dashboard(t, watch_for, interval_minutes)
    dashboard.run()
